#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test task Jacobian matrices against finite differences."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import BodyTask


class TestJacobians(unittest.TestCase):

    """
    Test task Jacobian matrices against finite differences.
    """

    def setUp(self, nb_configs=10, nb_dirs=2):
        np.random.seed(42)
        random_dq = 2.0 * np.random.random((nb_dirs, 6)) - 1.0
        l2norms = np.sqrt((random_dq * random_dq).sum(axis=1))
        random_dirs = random_dq / l2norms.reshape((nb_dirs, 1))
        random_q = (
            0.1 * np.pi * (2.0 * np.random.random((nb_configs, 6)) - 1.0)
        )
        robot = load_robot_description("ur3_description")
        self.assertEqual(robot.nq, 6)
        self.body = "ee_link"
        self.data = robot.data
        self.model = robot.model
        self.random_dirs = random_dirs
        self.random_q = random_q
        self.robot = robot

    def test_body_task(self, tol=1e-6):
        """
        Test BodyTask Jacobian matrix.

        Args:
            tol: Test tolerance.
        """
        task = BodyTask(self.body, position_cost=1.0, orientation_cost=1.0)
        task.set_target(pin.SE3.Random())

        def e(q):
            configuration = Configuration(self.model, self.data, q)
            return task.compute_error(configuration)

        def J(q):
            configuration = Configuration(self.model, self.data, q)
            return task.compute_jacobian(configuration)

        nq = self.robot.model.nq
        nv = self.robot.model.nv
        for q_0 in self.random_q:
            J_0 = J(q_0)
            e_0 = e(q_0)

            J_finite = np.empty((e_0.shape[0], nv))
            for i in range(nq):
                h = 0.000001
                e_i = np.eye(nq)[i]
                J_finite[:, i] = (e(q_0 + h * e_i) - e_0) / h

            self.assertLess(np.linalg.norm(J_0 + J_finite, ord=np.inf), tol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Build and solve the inverse kinematics problem."""

from typing import Iterable, Optional, Tuple

import numpy as np
import qpsolvers

from .configuration import Configuration
from .limits import compute_velocity_limits
from .tasks import Task


def __compute_qp_objective(
    configuration: Configuration, tasks: Iterable[Task], damping: float
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Compute the QP objective function.

    The Hessian matrix :math:`H` and linear vector :math:`c` define the QP
    objective function as:

    .. math::

        \frac{1}{2} \Delta q^T H \Delta q + c^T q

    The configuration displacement :math:`\Delta q` is the output of inverse
    kinematics (we divide it by :math:`\Delta t` to get a commanded velocity).

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks to fulfill at (weighted) best.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            `[cost]^2 / [tangent]` where `[tangent]` is "the" unit of robot
            velocities. Improves numerical stability, but larger values slow
            down all tasks.

    Returns:
        Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
        objective.
    """
    H = damping * configuration.tangent.eye
    c = configuration.tangent.zeros.copy()
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(configuration)
        H += H_task
        c += c_task
    return (H, c)


def __compute_qp_inequalities(
    configuration, dt
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    """Compute inequality constraints for the quadratic program.

    Args:
        configuration: Robot configuration to read kinematics from.
        dt: Integration timestep in [s].

    Returns:
        Pair :math:`(G, h)` of inequality matrix and vector.

    Notes:
        We trim comparisons to infinity (equivalently: big floats) because some
        solvers don't support it. See for instance
        https://github.com/tasts-robots/pink/issues/10.
    """
    v_max, v_min = compute_velocity_limits(configuration, dt)
    if v_max is None or v_min is None:
        return None, None

    bounded_tangent = configuration.model.bounded_tangent
    bounded_proj = bounded_tangent.projection_matrix
    G = np.vstack([bounded_proj, -bounded_proj])
    h = np.hstack([dt * v_max, -dt * v_min])
    return G, h


def build_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    damping: float = 1e-12,
) -> qpsolvers.Problem:
    r"""Build quadratic program from current configuration and tasks.

    This quadratic program is, in standard form:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{\Delta q}{\mbox{minimize}} &
                \frac{1}{2} {\Delta q}^T H {\Delta q} + c^T {\Delta q} \\
            \mbox{subject to}
                & G {\Delta q} \leq h
        \end{array}\end{split}

    where :math:`\Delta q` is a vector of joint displacements corresponding to
    the joint velocities :math:`v = {\Delta q} / {\mathrm{d}t}`.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.

    Returns:
        Quadratic program of the inverse kinematics problem.
    """
    configuration.check_limits()
    P, q = __compute_qp_objective(configuration, tasks, damping)
    G, h = __compute_qp_inequalities(configuration, dt)
    problem = qpsolvers.Problem(P, q, G, h)
    return problem


def solve_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    solver: str,
    damping: float = 1e-12,
    **kwargs,
) -> np.ndarray:
    r"""Compute a velocity tangent to the current robot configuration.

    The computed velocity satisfies at (weighted) best the set of kinematic
    tasks given in argument.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        solver: Backend quadratic programming (QP) solver.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.
        kwargs: Keyword arguments to forward to the backend QP solver.

    Returns:
        Velocity :math:`v` in tangent space.

    Raises:
        NotWithinConfigurationLimits: if the current configuration is not
            within limits.

    Note:
        Our Tikhonov damping is isotropic despite tangent velocities not being
        homogeneous. If it helps we can add a tangent-space scaling to damp the
        floating base differently from joint angular velocities.
    """
    configuration.check_limits()
    problem = build_ik(configuration, tasks, dt, damping)
    result = qpsolvers.solve_problem(problem, solver=solver, **kwargs)
    Delta_q = result.x
    assert Delta_q is not None
    v: np.ndarray = Delta_q / dt
    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization helpers."""

import pinocchio as pin


def start_meshcat_visualizer(robot: pin.RobotWrapper):
    """Open a MeshCat visualizer in a web browser.

    Args:
        robot: Robot wrapper with its initial data.

    Returns:
        MeshCat visualizer.
    """
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()
    return viz



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load a custom URDF."""

import os

import pinocchio as pin

try:
    import yourdfpy
except ImportError:
    yourdfpy = None

if __name__ == "__main__":
    urdf_path = os.path.join(
        os.path.dirname(__file__),
        "robots",
        "double_pendulum.urdf",
    )

    robot = pin.RobotWrapper.BuildFromURDF(
        filename=urdf_path,
        package_dirs=["."],
        root_joint=None,
    )
    print(f"URDF description successfully loaded in {robot}")

    if yourdfpy is None:
        print("If you ``pip install yourdfpy``, this example will display it.")
    else:  # yourdfpy is not None
        viz = yourdfpy.URDF.load(urdf_path)
        viz.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2023 Inria
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DRACO 3 humanoid standing on two feet and reaching with a hand."""

import os
import sys

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers
from loop_rate_limiters import RateLimiter

cwd = os.getcwd()
sys.path.append(cwd)
import pink
from pink import solve_ik
from pink.tasks import (
    BodyTask,
    JointCouplingTask,
    PostureTask,
)

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `pip install robot_descriptions`"
    )


class WavingPose:
    """Moving target to the wave the right hand."""

    def __init__(self, init: pin.SE3):
        """Initialize pose.

        Args:
            init: Initial transform from the wrist frame to the world frame.
        """
        self.init = init

    def at(self, t):
        """Get waving pose at a given time.

        Args:
            t: Time in seconds.
        """
        T = self.init.copy()
        R = T.rotation
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, 0.0, np.pi / 2))
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, -np.pi, 0.0))
        T.rotation = R
        T.translation[0] += 0.5
        T.translation[1] += 0.1 + 0.05 * np.sin(8.0 * t)
        T.translation[2] += 0.5
        return T


if __name__ == "__main__":
    robot = load_robot_description(
        "draco3_description", root_joint=pin.JointModelFreeFlyer()
    )

    # Initialize meschcat visualizer
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()

    # Set initial robot configuration
    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    viz.display(configuration.q)

    # Tasks initialization for IK
    left_foot_task = BodyTask(
        "l_foot_contact", position_cost=1.0, orientation_cost=1.0
    )
    pelvis_task = BodyTask(
        "torso_com_link", position_cost=1.0, orientation_cost=0.0
    )
    right_foot_task = BodyTask(
        "r_foot_contact", position_cost=1.0, orientation_cost=1.0
    )
    right_wrist_task = BodyTask(
        "r_hand_contact", position_cost=4.0, orientation_cost=4.0
    )
    posture_task = PostureTask(
        cost=1e-1,  # [cost] / [rad]
    )

    # Joint coupling task
    r_knee_holonomic_task = JointCouplingTask(
        ["r_knee_fe_jp", "r_knee_fe_jd"], [1.0, -1.0], 100.0, configuration
    )
    l_knee_holonomic_task = JointCouplingTask(
        ["l_knee_fe_jp", "l_knee_fe_jd"], [1.0, -1.0], 100.0, configuration
    )

    tasks = [
        left_foot_task,
        pelvis_task,
        right_foot_task,
        right_wrist_task,
        posture_task,
        l_knee_holonomic_task,
        r_knee_holonomic_task,
    ]

    # Task target specifications
    pelvis_pose = configuration.get_transform_body_to_world(
        "torso_com_link"
    ).copy()
    pelvis_pose.translation[0] += 0.05
    pelvis_task.set_target(pelvis_pose)

    transform_l_ankle_target_to_init = pin.SE3(
        np.eye(3), np.array([0.1, 0.0, 0.0])
    )
    transform_r_ankle_target_to_init = pin.SE3(
        np.eye(3), np.array([-0.1, 0.0, 0.0])
    )

    left_foot_task.set_target(
        configuration.get_transform_body_to_world("l_foot_contact")
        * transform_l_ankle_target_to_init
    )
    right_foot_task.set_target(
        configuration.get_transform_body_to_world("r_foot_contact")
        * transform_r_ankle_target_to_init
    )

    pelvis_task.set_target(
        configuration.get_transform_body_to_world("torso_com_link")
    )

    posture_task.set_target_from_configuration(configuration)

    right_wrist_pose = WavingPose(
        configuration.get_transform_body_to_world("r_hand_contact")
    )

    wrist_frame = viz.viewer["right_wrist_pose"]
    meshcat_shapes.frame(wrist_frame)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        right_wrist_task.set_target(right_wrist_pose.at(t))
        wrist_frame.set_transform(right_wrist_pose.at(t).np)

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test fixture for other library features."""

import unittest

import numpy as np
import pinocchio as pin
from pink.tasks.utils import body_minus, spatial_minus
from pink.utils import custom_configuration_vector
from robot_descriptions.loaders.pinocchio import load_robot_description


class TestUtils(unittest.TestCase):
    def test_custom_configuration_vector(self):
        """Check a custom configuration vector for Upkie.

        Assumes the left and right knees have joint indices respectively 8 and
        11 in the configuration vector.
        """
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        q = custom_configuration_vector(robot, left_knee=0.2, right_knee=-0.2)
        self.assertAlmostEqual(q[8], 0.2)
        self.assertAlmostEqual(q[11], -0.2)

    def test_minus(self):
        """Test Lie minus operators."""
        X = pin.SE3.Random()
        Y = pin.SE3.Random()
        self.assertTrue(np.allclose(Y, X * pin.exp6(body_minus(Y, X))))
        self.assertTrue(np.allclose(Y, pin.exp6(spatial_minus(Y, X)) * X))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Posture task implementation."""

from typing import Optional, Tuple

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from ..utils import get_root_joint_dim
from .exceptions import TargetNotSet
from .task import Task


class PostureTask(Task):
    r"""Regulate joint angles to a desired posture.

    A posture is a vector of actuated joint angles. Floating base coordinates
    are not affected by this task.

    Attributes:
        cost: joint angular error cost in
            :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
        target_q: Target vector in the configuration space.

    A posture task is typically used for regularization as it has a steady
    rank. For instance, when Upkie's legs are stretched and the Jacobian of its
    contact frames become singular, the posture task will drive the knees
    toward a preferred orientation.
    """

    cost: float
    target_q: Optional[np.ndarray]

    def __init__(self, cost: float) -> None:
        r"""Create task.

        Args:
            cost: joint angular error cost in
                :math:`[\mathrm{cost}] / [\mathrm{rad}]`.

        Note:
            We assume that the first seven coordinates of the configuration are
            for the floating base.
        """
        self.cost = cost
        self.target_q = None

    def set_target(self, target_q: np.ndarray) -> None:
        """Set target posture.

        Args:
            target_q: Target vector in the configuration space.
        """
        self.target_q = target_q.copy()

    def set_target_from_configuration(
        self, configuration: Configuration
    ) -> None:
        """Set target posture from a robot configuration.

        Args:
            configuration: Robot configuration.
        """
        self.set_target(configuration.q)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        """Compute posture task error.

        See :func:`Task.compute_error` for more context.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task error :math:`e(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet("no posture target")
        _, nv = get_root_joint_dim(configuration.model)
        return pin.difference(
            configuration.model, configuration.q, self.target_q
        )[nv:]

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute posture task Jacobian.

        The task Jacobian is the identity :math:`I_{n_v} \in \mathbb{R}^{n_v
        \times n_v}`, with :math:`n_v` the dimension of the robot's tangent
        space, so that the task dynamics are:

        .. math::

            J(q) \Delta q = \Delta_q = \alpha (q^* - q)

        See :func:`Task.compute_jacobian` for more context.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task Jacobian :math:`J(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet("no posture target")
        _, nv = get_root_joint_dim(configuration.model)
        return configuration.tangent.eye[nv:, :]

    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the matrix-vector pair :math:`(H, c)` of the QP objective.

        This pair is such that the contribution of the task to the QP objective
        of the IK is:

        .. math::

            \| J \Delta q - \alpha e \|_{W}^2
            = \frac{1}{2} \Delta q^T H \Delta q + c^T q

        The weight matrix :math:`W \in \mathbb{R}^{n \times n}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is :math:`[\mathrm{cost}]^2`. The configuration
        displacement :math:`\Delta q` is the output of inverse kinematics (we
        divide it by :math:`\Delta t` to get a commanded velocity).

        Args:
            configuration: Robot configuration.

        Returns:
            Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
            objective.
        """
        jacobian = self.compute_jacobian(configuration)
        gain_error = self.gain * self.compute_error(configuration)
        weighted_jacobian = self.cost * jacobian  # [cost]
        weighted_error = self.cost * gain_error  # [cost]
        H = weighted_jacobian.T @ weighted_jacobian
        c = -weighted_error.T @ weighted_jacobian
        return (H, c)

    def __repr__(self):
        """Human-readable representation of the task."""
        return f"PostureTask(cost={self.cost}, gain={self.gain})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint limits implemented as inequality constraints.

You can check out a writeup of this formulation in `this post
<https://scaron.info/robot-locomotion/inverse-kinematics.html#inequality-constraints>`__.
"""

from typing import Optional, Tuple

import numpy as np
import pinocchio as pin

from .bounded_tangent import BoundedTangent
from .configuration import Configuration


def compute_velocity_limits(
    configuration: Configuration,
    dt: float,
    config_limit_gain: float = 0.5,
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    r"""Compute the configuration-dependent velocity limits.

    Those limits are defined by:

    .. math::

        v_{min}(q) \leq v \leq v_{max}(q)

    where :math:`q \in {\cal C}` is the robot's configuration and :math:`v
    \in T_q({\cal C})` is the velocity in the tangent space at :math:`q`.
    These velocity limits combine:

    - Configuration-independent joint velocity limits
      :math:`|\dot{q}| \leq \dot{q}_{lim}`.
    - Time-derivatives of the configuration limit
      :math:`q_{min} \leq q \leq q_{max}`.

    Args:
        configuration: Robot configuration to read kinematics from.
        dt: Integration timestep in [s].
        config_limit_gain: gain between 0 and 1 to steer away from
            configuration limits. It is described in "Real-time prioritized
            kinematic control under inequality constraints for redundant
            manipulators" (Kanoun, 2012). More details in `this writeup
            <https://scaron.info/teaching/inverse-kinematics.html>`__.

    Returns:
        Pair :math:`(v_{max}(q), v_{min}(q))` of velocity lower and upper
        bounds, or ``(None, None)`` if there is no velocity limit.
    """
    assert 0.0 < config_limit_gain <= 1.0
    model = configuration.model
    bounded_tangent: BoundedTangent = model.bounded_tangent
    if bounded_tangent.velocity_limit is None:
        return None, None

    # Velocity limits from configuration bounds
    Delta_q_max = bounded_tangent.project(
        pin.difference(
            model,
            configuration.q,
            model.upperPositionLimit,
        )
    )
    Delta_q_min = bounded_tangent.project(
        pin.difference(
            model,
            configuration.q,
            model.lowerPositionLimit,
        )
    )

    # Intersect with velocity limits from URDF
    v_max = np.minimum(
        bounded_tangent.velocity_limit,
        config_limit_gain * Delta_q_max / dt,
    )
    v_min = np.maximum(
        -bounded_tangent.velocity_limit,
        config_limit_gain * Delta_q_min / dt,
    )

    return v_max, v_min



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration type.

Pink uses `Pinocchio <https://github.com/stack-of-tasks/pinocchio>`__ for
forward kinematics. It adds to it a :class:`Configuration` type to indicate
that forward kinematics functions have been run, indicating that frame
transforms and frame Jacobians used for IK can be queried from the robot's
data.
"""

import numpy as np
import pinocchio as pin

from .bounded_tangent import BoundedTangent
from .exceptions import BodyNotFound, NotWithinConfigurationLimits
from .utils import VectorSpace, get_root_joint_dim


class Configuration:
    """Type indicating that configuration-dependent quantities are available.

    In Pink, this type enables access to frame transforms and frame Jacobians.
    We rely on typing to make sure the proper forward kinematics functions have
    been called beforehand. In Pinocchio, these functions are:

    - ``pin.computeJointJacobians(model, data, configuration)``
    - ``pin.updateFramePlacements(model, data)``

    The former computes the full model Jacobian into ``data.J``. (It also
    computes forward kinematics, so there is no need to further call
    ``pin.forwardKinematics(model, data, configuration)``.) The latter updates
    frame placements.

    Notes:
        This class is meant to be used as a subclass of pin.RobotWrapper, not
        wrap it. However, right now pin.RobotWrapper does not have a shallow
        copy constructor. TODO(scaron): bring it up upstream.

    Attributes:
        data: Data corresponding to :data:`Configuration.model`.
        model: Kinodynamic model.
        q: Configuration vector for the robot model.
    """

    data: pin.Data
    model: pin.Model
    q: np.ndarray

    def __init__(
        self,
        model: pin.Model,
        data: pin.Data,
        q: np.ndarray,
        copy_data=True,
        forward_kinematics=True,
    ):
        """Initialize configuration.

        Args:
            model: Kinodynamic model.
            data: Data corresponding to the model.
            q: Configuration vector.
            copy_data: If true (default), work on an internal copy of the input
                data. Otherwise, work on the input data directly.
            forward_kinematics: If true (default), compute forward kinematics
                from the q into the internal data.

        Notes:
            Configurations copy data and run forward kinematics by default so
            that they are less error-prone for newcomers. You can avoid copies
            or forward kinematics (e.g. if it is already computed by the
            caller) using constructor parameters.
        """
        if not hasattr(model, "tangent"):
            model.tangent = VectorSpace(model.nv)
        if not hasattr(model, "bounded_tangent"):
            model.bounded_tangent = BoundedTangent(model)
        q_readonly = q.copy()
        q_readonly.setflags(write=False)
        self.data = data.copy() if copy_data else data
        self.model = model
        self.q = q_readonly
        self.tangent = model.tangent
        #
        if forward_kinematics:
            self.update()

    def update(self) -> None:
        """Run forward kinematics from the configuration."""
        pin.computeJointJacobians(self.model, self.data, self.q)
        pin.updateFramePlacements(self.model, self.data)

    def check_limits(self, tol: float = 1e-6) -> None:
        """Check that the current configuration is within limits.

        Args:
            tol: Tolerance in radians.

        Raises:
            NotWithinConfigurationLimits: if the current configuration is
                within limits.
        """
        q_max = self.model.upperPositionLimit
        q_min = self.model.lowerPositionLimit
        root_nq, _ = get_root_joint_dim(self.model)
        for i in range(root_nq, self.model.nq):
            if q_max[i] <= q_min[i] + tol:  # no limit
                continue
            if self.q[i] < q_min[i] - tol or self.q[i] > q_max[i] + tol:
                raise NotWithinConfigurationLimits(
                    i,
                    self.q[i],
                    q_min[i],
                    q_max[i],
                )

    def get_body_jacobian(self, body: str) -> np.ndarray:
        r"""Compute the Jacobian matrix of the body velocity.

        This matrix :math:`{}_B J_{WB}` is related to the body velocity
        :math:`{}_B v_{WB}` by:

        .. math::

            {}_B v_{WB} = {}_B J_{WB} \dot{q}

        Args:
            body: Body frame name, typically the link name from the URDF.

        Returns:
            Jacobian :math:`{}_B J_{WB}` of the body twist.

        When the robot model includes a floating base
        (pin.JointModelFreeFlyer), the configuration vector :math:`q` consists
        of:

        - ``q[0:3]``: position in [m] of the floating base in the inertial
          frame, formatted as :math:`[p_x, p_y, p_z]`.
        - ``q[3:7]``: unit quaternion for the orientation of the floating base
          in the inertial frame, formatted as :math:`[q_x, q_y, q_z, q_w]`.
        - ``q[7:]``: joint angles in [rad].
        """
        if not self.model.existBodyName(body):
            raise BodyNotFound(f"body {body} does not exist")
        body_id = self.model.getBodyId(body)
        J: np.ndarray = pin.getFrameJacobian(
            self.model, self.data, body_id, pin.ReferenceFrame.LOCAL
        )
        return J

    def get_transform_body_to_world(self, body: str) -> pin.SE3:
        """Get the pose of a body frame in the current configuration.

        Args:
            body: Body frame name, typically the link name from the URDF.

        Returns:
            Current transform from body frame to world frame.

        Raises:
            KeyError: if the body name is not found in the robot model.
        """
        body_id = self.model.getBodyId(body)
        try:
            return self.data.oMf[body_id].copy()
        except IndexError as index_error:
            raise KeyError(
                f'Body "{body}" not found in robot model'
            ) from index_error

    def integrate(self, velocity, dt) -> np.ndarray:
        """Integrate a velocity starting from the current configuration.

        Args:
            velocity: Velocity in tangent space.
            dt: Integration duration in [s].

        Returns:
            New configuration vector after integration.
        """
        return pin.integrate(self.model, self.q, velocity * dt)

    def integrate_inplace(self, velocity, dt) -> None:
        """Integrate a velocity starting from the current configuration.

        Args:
            velocity: Velocity in tangent space.
            dt: Integration duration in [s].
        """
        self.q = pin.integrate(self.model, self.q, velocity * dt)
        self.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test velocity and actuated-joint limits.
"""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.limits import compute_velocity_limits


class TestLimits(unittest.TestCase):
    def setUp(self):
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.data = robot.data
        self.model = robot.model
        self.robot = robot

    def test_limit_dimension(self):
        """Velocity limit dimension is the number of bounded joints."""
        configuration = Configuration(self.model, self.data, self.robot.q0)
        v_max, v_min = compute_velocity_limits(configuration, dt=1e-3)
        self.assertEqual(v_max.shape, (self.model.bounded_tangent.dim,))
        self.assertEqual(v_min.shape, (self.model.bounded_tangent.dim,))

    def test_model_with_no_joint_limit(self):
        """Model with no joint limit has no velocity-limit vector."""
        model = pin.Model()
        model.addJoint(
            0, pin.JointModelSpherical(), pin.SE3.Identity(), "spherical"
        )
        robot = pin.RobotWrapper(model=model)
        configuration = Configuration(robot.model, robot.data, robot.q0)
        v_max, v_min = compute_velocity_limits(configuration, dt=1e-3)
        self.assertIsNone(v_max)
        self.assertIsNone(v_min)

    def test_model_with_limitless_joint(self):
        """Same as previous test, but the joint has a limit set to zero."""
        model = pin.Model()
        model.addJoint(
            0,
            pin.JointModelRevoluteUnaligned(),
            pin.SE3.Identity(),
            "revolute",
            max_effort=np.array([0.0]),
            max_velocity=np.array([0.0]),
            min_config=np.array([0.0]),
            max_config=np.array([0.0]),
        )
        robot = pin.RobotWrapper(model=model)
        configuration = Configuration(robot.model, robot.data, robot.q0)
        v_max, v_min = compute_velocity_limits(configuration, dt=1e-3)
        self.assertIsNone(v_max)
        self.assertIsNone(v_min)

    def test_forward_velocity_limit(self):
        """Velocity limits have no effect far from configuration limits.

        When we are far away from configuration limits, the velocity limit is
        simply the configuration-agnostic one from the robot.
        """
        configuration = Configuration(self.model, self.data, self.robot.q0)
        v_max, v_min = compute_velocity_limits(configuration, dt=1e-3)
        v_lim = configuration.model.bounded_tangent.velocity_limit
        tol = 1e-10
        self.assertLess(np.max(v_max - v_lim), tol)
        self.assertLess(np.max(-v_lim - v_min), tol)

    def test_configuration_limit_repulsion(self):
        """Velocities are scaled down when close to a configuration limit."""
        dt = 1e-3  # [s]
        configuration = Configuration(self.model, self.data, self.robot.q0)
        slack_vel = 5.5e-4 * configuration.tangent.ones
        bounded_slack_vel = slack_vel[self.model.bounded_tangent.indices]
        self.model.upperPositionLimit = configuration.integrate(slack_vel, dt)
        v_max, v_min = compute_velocity_limits(
            configuration, dt, config_limit_gain=0.5
        )
        tol = 1e-10
        self.assertLess(
            np.max(v_max - self.model.bounded_tangent.velocity_limit), tol
        )
        self.assertLess(np.max(v_max - bounded_slack_vel), tol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test submodels.
"""

import unittest

import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink.bounded_tangent import BoundedTangent
from pink.utils import VectorSpace


class TestBoundedTangent(unittest.TestCase):

    """
    Test fixture for bounded tangent subspace.
    """

    def setUp(self):
        """
        Set test fixture up.
        """
        model = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        ).model
        self.bounded_tangent = BoundedTangent(model)
        self.model = model
        self.tangent = VectorSpace(model.nv)

    def test_tangent(self):
        """
        Check dimensions of regular tangent space.
        """
        nv = self.model.nv
        self.assertEqual(self.tangent.eye.shape, (nv, nv))
        self.assertEqual(self.tangent.ones.shape, (nv,))
        self.assertEqual(self.tangent.zeros.shape, (nv,))

    def test_bounded_tangent(self):
        """
        Check dimensions in bounded tangent space.
        """
        for joint in self.bounded_tangent.joints:
            self.assertGreaterEqual(joint.idx_q, 0)
            self.assertGreaterEqual(joint.idx_v, 0)
        nb = len(self.bounded_tangent.joints)  # those are only revolute joints
        nv = self.model.nv
        self.assertEqual(self.bounded_tangent.dim, nb)
        self.assertEqual(self.bounded_tangent.nv, nv)
        self.assertEqual(
            self.bounded_tangent.projection_matrix.shape, (nb, nv)
        )

    def test_unbounded_tangent(self):
        """
        Check that unbounded models don't fail.
        """
        empty_model = pin.Model()
        empty_bounded = BoundedTangent(empty_model)
        self.assertEqual(empty_bounded.dim, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Subset of bounded joints associated with a robot model."""

from typing import List, Optional

import numpy as np
import pinocchio as pin

from .utils import VectorSpace


class BoundedTangent(VectorSpace):
    """Subspace of the tangent space restricted to bounded joints.

    Attributes:
        nv: Dimension of the full tangent space.
    """

    indices: np.ndarray
    joints: list
    nv: int
    projection_matrix: Optional[np.ndarray]
    velocity_limit: Optional[np.ndarray]

    def __init__(self, model: pin.Model):
        """Bounded joints in a robot model.

        Args:
            model: robot model.

        Returns:
            List of bounded joints.
        """
        has_configuration_limit = np.logical_and(
            model.upperPositionLimit < 1e20,
            model.upperPositionLimit > model.lowerPositionLimit + 1e-10,
        )

        joints = [
            joint
            for joint in model.joints
            if joint.idx_q >= 0
            and has_configuration_limit[
                slice(joint.idx_q, joint.idx_q + joint.nq)
            ].all()
        ]

        index_list: List[int] = []
        for joint in joints:
            index_list.extend(range(joint.idx_v, joint.idx_v + joint.nv))
        indices = np.array(index_list)
        indices.setflags(write=False)

        dim = len(indices)
        super().__init__(dim)
        projection_matrix = np.eye(model.nv)[indices] if dim > 0 else None

        self.indices = indices
        self.joints = joints
        self.nv = model.nv
        self.projection_matrix = projection_matrix
        self.velocity_limit = (
            model.velocityLimit[indices] if len(joints) > 0 else None
        )

    def project(self, v: np.ndarray) -> np.ndarray:
        """Project a tangent vector to the bounded tangent subspace.

        Args:
            v: Vector from the original space.
        """
        assert v.shape == (self.nv,), "Dimension mismatch"
        return v[self.indices]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lie-algebra utility functions."""

import numpy as np
import pinocchio as pin


def spatial_minus(Y: pin.SE3, X: pin.SE3) -> np.ndarray:
    r"""Compute the left minus :math:`Y \ominus_0 X`.

    The left minus operator is defined by:

    .. math::

        Y \ominus_0 X = \log(Y \cdot X^{-1})

    This operator allows us to think about orientation "differences" as
    similarly as possible to position differences, but mind the frames! If we
    denote by :math:`Y = T_{0y}` and :math:`X = T_{0x}` the two transforms,
    from respectively frames :math:`y` and :math:`x` to the inertial frame
    :math:`0`, the twist resulting from a left minus is expressed in the
    inertial frame:

    .. math::

        {}_0 \xi_{0x} = Y \ominus_0 X = \log(T_{0y} T_{x0})

    A twist like :math:`{}_0 \xi_{0x}` in the inertial frame is called a
    spatial motion vector.

    Args:
        Y: Transform :math:`Y = T_{0y}` on the left-hand side of the operator.
        X: Transform :math:`X = T_{0x}` on the right-hand side of the operator.

    Returns:
        Spatial motion vector resulting from the difference :math:`\ominus_0`
        between :math:`Y` and :math:`X`.

    Notes:
        The `micro Lie theory <https://arxiv.org/abs/1812.01537>`_ describes
        the difference between the left and right minus operators.
    """
    spatial_twist: np.ndarray = pin.log(Y.act(X.inverse())).vector
    return spatial_twist


def body_minus(Y: pin.SE3, X: pin.SE3) -> np.ndarray:
    r"""Compute the right minus :math:`Y \ominus X`.

    The right minus operator is defined by:

    .. math::

        Y \ominus X = \log(X^{-1} \cdot Y)

    This operator allows us to think about orientation "differences" as
    similarly as possible to position differences, but mind the frames! If we
    denote by :math:`Y = T_{0y}` and :math:`X = T_{0x}` the two transforms,
    from respectively frames :math:`y` and :math:`x` to the inertial frame
    :math:`0`, the twist resulting from a right minus is expressed in the
    local frame :math:`x`:

    .. math::

        {}_x \xi_{0x} = Y \ominus X = \log(T_{x0} T{0y}) = \log(T_{xy})

    A twist like :math:`{}_x \xi_{0x}` in the local frame :math:`x` is called a
    body motion vector.

    Args:
        Y: Transform :math:`Y = T_{0y}` on the left-hand side of the operator.
        X: Transform :math:`X = T_{0x}` on the right-hand side of the operator.

    Returns:
        Body motion vector resulting from the difference :math:`\ominus_0`
        between :math:`Y` and :math:`X`.

    Notes:
        - `Body motion vector
          <https://scaron.info/robotics/screw-theory.html#body-screws>`_ are
          also known as body screws in screw theory.
        - The `micro Lie theory <https://arxiv.org/abs/1812.01537>`_ describes
          the difference between the left and right minus operators.
    """
    body_twist: np.ndarray = pin.log(X.actInv(Y)).vector
    return body_twist



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint coupling task implementation."""

from typing import Sequence

import numpy as np

from ..configuration import Configuration
from ..utils import get_joint_idx
from .linear_holonomic_task import LinearHolonomicTask


class JointCouplingTask(LinearHolonomicTask):
    r"""Coupling between two revolute joints.

    Note:
        This task only considers a 1-Dimensional task.
    """

    def __init__(
        self,
        joint_name_list: Sequence[str],
        ratios: Sequence[float],
        cost: float,
        configuration: Configuration,
    ) -> None:
        r"""Compute Jacobian matrix of a linear holonomic constraint.

        Args:
            joint_name_list: a list of joint names consisting of a linear
                holonomic constraint.
            ratios: a list of ratios that each joint takes in a linear
                holonomic constraint.
            cost: Cost of a linear holonomic task.
            configuration: Robot configuration.
        """
        assert len(joint_name_list) == len(ratios)
        A = np.zeros((1, configuration.model.nv))
        for joint, ratio in zip(joint_name_list, ratios):
            _, joint_v_idx = get_joint_idx(configuration.model, joint)
            A[:, joint_v_idx] = ratio
        super().__init__(A, cost)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""All kinematic tasks derive from the :class:`Task` base class.

The formalism used in this implementation is written down in `this note on
task-based inverse kinematics
<https://scaron.info/robotics/inverse-kinematics.html>`_. As of February 2022
it hasn't been updated with the proper dimensional analysis, but the core
concepts and notations are there.
"""

import abc
from typing import Tuple

import numpy as np

from ..configuration import Configuration


class Task(abc.ABC):
    r"""Abstract base class for kinematic tasks.

    Attributes:
        gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
            filtering. Defaults to 1.0 (no filtering) for dead-beat control.
    """

    gain: float = 1.0

    @abc.abstractmethod
    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task error function.

        The error function :math:`e(q) \in \mathbb{R}^{k}` is the quantity that
        the task aims to drive to zero (:math:`k` is the dimension of the
        task). It appears in the first-order task dynamics:

        .. math::

            J(q) \Delta q = \alpha e(q)

        The Jacobian matrix :math:`J(q) \in \mathbb{R}^{k \times n_v}`,
        with :math:`n_v` the dimension of the robot's tangent space, is
        computed by :func:`Task.compute_jacobian`, while the configuration
        displacement :math:`\\Delta q` is the output of inverse kinematics. The
        error vector :math:`e(q)` is multiplied by the task gain :math:`\alpha
        \in [0, 1]`. The gain is usually 1 for dead-beat control (*i.e.*
        converge as fast as possible), but it can also be lower for some extra
        low-pass filtering.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error vector :math:`e(q)`.
        """

    @abc.abstractmethod
    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task Jacobian at a given configuration.

        The task Jacobian :math:`J(q) \in \mathbb{R}^{k \times n_v}` appears in
        the first-order task dynamics:

        .. math::

            J(q) \Delta q = \alpha e(q)

        The error :math:`e(q) \in \mathbb{R}^{k \times n_v}`, with :math:`k`
        the dimension of the task and :math:`n_v` the dimension of the robot's
        tangent space, is computed by :func:`Task.compute_error`, while the
        configuration displacement :math:`\\Delta q` is the output of inverse
        kinematics.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task Jacobian :math:`J(q)`.
        """

    @abc.abstractmethod
    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the matrix-vector pair :math:`(H, c)` of the QP objective.

        This pair is such that the contribution of the task to the QP objective
        of the IK is:

        .. math::

            \| J \Delta q - \alpha e \|_{W}^2 = \frac{1}{2} \Delta q^T H
            \Delta q + c^T q

        The weight matrix :math:`W \in \mathbb{R}^{k \times k}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is [cost]^2. The configuration displacement :math:`\Delta
        q` is the output of inverse kinematics (we divide it by :math:`\Delta
        t` to get a commanded velocity).

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Pair :math:`(H(q), c(q))` of Hessian matrix and linear vector of
            the QP objective.
        """

    def __repr__(self):
        """Human-readable representation of the task."""
        return f"Task(gain={self.gain})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Linear holonomic task :math:`A q = b`."""

from typing import Sequence, Tuple, Union

import numpy as np

from ..configuration import Configuration
from .exceptions import TaskJacobianNotSet
from .task import Task


class LinearHolonomicTask(Task):
    r"""Linear holonomic task :math:`A q = 0`.

    Attributes:
        A: matrix that defines the task:

            .. math::
                e(q) = A q
                \dot{e}(q) := A \dot{q}

            where :math: `e(q) \in \mathbb{R}^{k}` is the quantity that the
            task aims to drive to zero (:math:`k` is the dimension of the
            task).
        cost: joint angular error cost in
            :math:`[\mathrm{cost}] / [\mathrm{rad}]`.

    Note:
        A linear holonomic task is typically used for a robot
        that has mechanical constraint (e.g., closed loop kinematics).
        Floating base coordinates are not affected by this task.
    """

    A: np.ndarray
    cost: Union[float, Sequence[float]]

    def __init__(
        self, A: np.ndarray, cost: Union[float, Sequence[float]]
    ) -> None:
        r"""Create task.

        Args:
            A: Jacobian matrix of the task.
            cost: joint angular error cost in
                :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
        """
        assert A.shape[0] == 1 if isinstance(cost, float) else len(cost)
        self.A = A
        self.cost = cost

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task error function.

        The error function :math:`e(q) \in \mathbb{R}^{k}` is the quantity that
        the task aims to drive to zero (:math:`k` is the dimension of the
        task). It appears in the first-order task dynamics:

        .. math::

            J(q) \Delta q = \alpha e(q)

        The Jacobian matrix :math:`J(q) \in \mathbb{R}^{k \times n_v}`,
        with :math:`n_v` the dimension of the robot's tangent space, is
        computed by :func:`Task.compute_jacobian`, while the configuration
        displacement :math:`\\Delta q` is the output of inverse kinematics. The
        error vector :math:`e(q)` is multiplied by the task gain :math:`\alpha
        \in [0, 1]`. The gain is usually 1 for dead-beat control (*i.e.*
        converge as fast as possible), but it can also be lower for some extra
        low-pass filtering.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error vector :math:`e(q)`.
        """
        if self.A.shape[1] != configuration.model.nv:
            raise TaskJacobianNotSet(
                "task Jacobian dimension is not set properly"
            )
        return np.zeros(self.A.shape[0])

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task Jacobian at a given configuration.

        The task Jacobian :math:`J(q) \in \mathbb{R}^{k \times n_v}` appears in
        the first-order task dynamics:

        .. math::

            J(q) \Delta q = \alpha e(q)

        The error :math:`e(q) \in \mathbb{R}^{k \times n_v}`, with :math:`k`
        the dimension of the task and :math:`n_v` the dimension of the robot's
        tangent space, is computed by :func:`Task.compute_error`, while the
        configuration displacement :math:`\\Delta q` is the output of inverse
        kinematics.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task Jacobian :math:`J(q)`.
        """
        if self.A.shape[1] != configuration.model.nv:
            raise TaskJacobianNotSet(
                "task Jacobian dimension is not set properly"
            )
        return self.A

    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the matrix-vector pair :math:`(H, c)` of the QP objective.

        This pair is such that the contribution of the task to the QP objective
        of the IK is:

        .. math::

            \| J \Delta q - \alpha e \|_{W}^2
            = \frac{1}{2} \Delta q^T H \Delta q + c^T q

        The weight matrix :math:`W \in \mathbb{R}^{n \times n}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is :math:`[\mathrm{cost}]^2`. The configuration
        displacement :math:`\Delta q` is the output of inverse kinematics (we
        divide it by :math:`\Delta t` to get a commanded velocity).

        Args:
            configuration: Robot configuration.

        Returns:
            Pair :math:`(H)` of Hessian matrix of the QP objective.
        """
        jacobian = self.compute_jacobian(configuration)
        gain_error = self.gain * self.compute_error(configuration)
        weight = np.diag(
            [self.cost] * jacobian.shape[0]
            if isinstance(self.cost, float)
            else self.cost
        )
        weighted_jacobian = np.dot(weight, jacobian)  # [cost]
        weighted_error = np.dot(weight, gain_error)  # [cost]
        H = weighted_jacobian.T @ weighted_jacobian
        c = -weighted_error.T @ weighted_jacobian
        return (H, c)

    def __repr__(self):
        """Human-readable representation of the task."""
        return f"LinearHolonomicTask(cost={self.cost}, gain={self.gain})"
